#!/usr/bin/env python

import sys
import xml.dom.minidom
from urllib.parse import quote


class dom_job:
    """infiles[inds]=[file1,file2...]
    outfiles = [file1,file2...]
    command  - script that will be executed on the grid
    prepend  - list of (option,value) prepended to output file name
    forward  - list of (option,value) forwarded to the grid job
    """

    def __init__(s, domjob=None, primaryds=None, defaultcmd=None, defaultout=[]):
        """Loads <job></job> from xml file.
        If primaryds is set, makes sure it is present in job spec"""
        s.infiles = {}
        s.outfiles = []
        s.command = defaultcmd
        s.prepend = []
        s.forward = []
        if not domjob:
            return
        # script executed on the grid node for this job
        if len(domjob.getElementsByTagName("command")) > 0:
            s.command = dom_parser.text(domjob.getElementsByTagName("command")[0])
        # input files
        for inds in domjob.getElementsByTagName("inds"):
            name = dom_parser.text(inds.getElementsByTagName("name")[0])
            files = inds.getElementsByTagName("file")
            if len(files) == 0:
                continue
            s.infiles[name] = []
            for file in files:
                s.infiles[name].append(dom_parser.text(file))
        if primaryds and primaryds not in s.infiles.keys():
            print(f"ERROR: primaryds={primaryds} must be present in each job")
            sys.exit(0)
        # output files (also, drop duplicates within this job)
        outfiles = set(defaultout)
        [outfiles.add(dom_parser.text(v)) for v in domjob.getElementsByTagName("output")]
        s.outfiles = list(outfiles)
        # gearing options
        for o in domjob.getElementsByTagName("option"):
            name = o.attributes["name"].value
            value = dom_parser.text(o)
            prepend = dom_parser.true(o.attributes["prepend"].value)
            forward = dom_parser.true(o.attributes["forward"].value)
            if prepend:
                s.prepend.append((name, value))
            if forward:
                s.forward.append((name, value))

    def to_dom(s):
        """Converts this job to a dom tree branch"""
        x = xml.dom.minidom.Document()
        job = x.createElement("job")
        for inds in s.infiles.keys():
            job.appendChild(x.createElement("inds"))
            job.childNodes[-1].appendChild(x.createElement("name"))
            job.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(inds))
            for file in s.infiles[inds]:
                job.childNodes[-1].appendChild(x.createElement("file"))
                job.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(file))
        for outfile in s.outfiles:
            job.appendChild(x.createElement("output"))
            job.childNodes[-1].appendChild(x.createTextNode(outfile))
        if s.command:
            job.appendChild(x.createElement("command"))
            job.childNodes[-1].appendChild(x.createTextNode(s.command))
        for option in s.prepend + list(set(s.prepend + s.forward) - set(s.prepend)):
            job.appendChild(x.createElement("option"))
            job.childNodes[-1].setAttribute("name", str(option[0]))
            if option in s.forward:
                job.childNodes[-1].setAttribute("forward", "true")
            else:
                job.childNodes[-1].setAttribute("forward", "false")
            if option in s.prepend:
                job.childNodes[-1].setAttribute("prepend", "true")
            else:
                job.childNodes[-1].setAttribute("prepend", "false")
            job.childNodes[-1].appendChild(x.createTextNode(str(option[1])))
        return job

    def files_in_DS(s, DS):
        """Returns a list of files used in a given job in a given dataset"""
        if DS in s.infiles:
            return s.infiles[DS]
        else:
            return []

    def forward_opts(s):
        """passable string of forward options"""
        return " ".join([f"{v[0]}={v[1]}" for v in s.forward])

    def prepend_string(s):
        """a tag string prepended to output files"""
        return "_".join([f"{v[0]}{v[1]}" for v in s.prepend])

    def exec_string(s):
        """exec string for prun.
        If user requested to run script run.sh (via <command>run.sh</command>), it will return
        opt1=value1 opt2=value2 opt3=value3 run.sh
        This way, all options will be set inside run.sh
        """
        return f"{s.forward_opts()} {s.command}"

    def exec_string_enc(s):
        """exec string for prun.
        If user requested to run script run.sh (via <command>run.sh</command>), it will return
        opt1=value1 opt2=value2 opt3=value3 run.sh
        This way, all options will be set inside run.sh
        """
        comStr = f"{s.forward_opts()} {s.command}"
        return quote(comStr)

    def get_outmap_str(s, outMap):
        """return mapping of original and new filenames"""
        newMap = {}
        for oldLFN, fileSpec in outMap.items():
            newMap[oldLFN] = str(fileSpec.lfn)
        return str(newMap)

    def outputs_list(s, prepend=False):
        """python list with finalized output file names"""
        if prepend and s.prepend_string():
            return [s.prepend_string() + "." + o for o in s.outfiles]
        else:
            return [o for o in s.outfiles]

    def outputs(s, prepend=False):
        """Comma-separated list of output files accepted by prun"""
        return ",".join(s.outputs_list(prepend))


class dom_parser:
    def __init__(s, fname=None, xmlStr=None):
        """creates a dom object out of a text file (if provided)"""
        s.fname = fname
        s.dom = None
        s.title = None
        s.tag = None
        s.command = None
        s.outds = None
        s.inds = {}
        s.global_outfiles = []
        s.jobs = []
        s.primaryds = None
        if fname:
            s.dom = xml.dom.minidom.parse(fname)
            s.parse()
            s.check()
        if xmlStr is not None:
            s.dom = xml.dom.minidom.parseString(xmlStr)
            s.parse()
            s.check()

    @staticmethod
    def break_regex(v, N=100):
        """breaks up a very long regex into a comma-separeted list of filters"""
        _REGEXLIM = 2**15 - 1000
        spl = v.split("|")
        res = []
        for i in range(0, len(spl), N):
            piece = "|".join(spl[i : i + N])
            assert len(piece) < _REGEXLIM, "Input dataset contains very long filenames. You must reduce parameter N in break_regex()"
            res.append(piece)
        return ",".join(res)

    @staticmethod
    def true(v):
        """define True"""
        return v in ("1", "true", "True", "TRUE", "yes", "Yes", "YES")

    @staticmethod
    def text(pnode):
        """extracts the value stored in the node"""
        rc = []
        for node in pnode.childNodes:
            if node.nodeType == node.TEXT_NODE:
                rc.append(str(node.data).strip())
        return "".join(rc)

    def parse(s):
        """loads submission configuration from an xml file"""
        try:
            # general settings
            if len(s.dom.getElementsByTagName("title")) > 0:
                s.title = dom_parser.text(s.dom.getElementsByTagName("title")[0])
            else:
                s.title = "Default title"
            if len(s.dom.getElementsByTagName("tag")) > 0:
                s.tag = dom_parser.text(s.dom.getElementsByTagName("tag")[0])
            else:
                s.tag = "default_tag"
            s.command = None  # can be overridden in subjobs
            for elm in s.dom.getElementsByTagName("submission")[0].childNodes:
                if elm.nodeName != "command":
                    continue
                s.command = dom_parser.text(elm)
                break
            s.global_outfiles = []  # subjobs can append *additional* outputs
            for elm in s.dom.getElementsByTagName("submission")[0].childNodes:
                if elm.nodeName != "output":
                    continue
                s.global_outfiles.append(dom_parser.text(elm))
            s.outds = dom_parser.text(s.dom.getElementsByTagName("outds")[0])
            # declaration of all input datasets
            primarydss = []
            for elm in s.dom.getElementsByTagName("submission")[0].childNodes:
                if elm.nodeName != "inds":
                    continue
                if "primary" in elm.attributes.keys() and dom_parser.true(elm.attributes["primary"].value):
                    primary = True
                else:
                    primary = False
                stream = dom_parser.text(elm.getElementsByTagName("stream")[0])
                name = dom_parser.text(elm.getElementsByTagName("name")[0])
                s.inds[name] = stream
                if primary:
                    primarydss.append(name)
            # see if one of the input datasets was explicitly labeled as inDS
            if len(primarydss) == 1:
                s.primaryds = primarydss[0]
            else:
                s.primaryds = None
            for job in s.dom.getElementsByTagName("job"):
                s.jobs.append(dom_job(job, primaryds=s.primaryds, defaultcmd=s.command, defaultout=s.global_outfiles))
        except Exception:
            print("ERROR: failed to parse" + " " + s.fname)
            raise

    def to_dom(s):
        """Converts this submission to a dom tree branch"""
        x = xml.dom.minidom.Document()
        submission = x.createElement("submission")
        if s.title:
            submission.appendChild(x.createElement("title"))
            submission.childNodes[-1].appendChild(x.createTextNode(s.title))
        if s.tag:
            submission.appendChild(x.createElement("tag"))
            submission.childNodes[-1].appendChild(x.createTextNode(s.tag))
        for name, stream in s.inds.items():
            submission.appendChild(x.createElement("inds"))
            if name == s.primaryds:
                submission.childNodes[-1].setAttribute("primary", "true")
            else:
                submission.childNodes[-1].setAttribute("primary", "false")
            submission.childNodes[-1].appendChild(x.createElement("stream"))
            submission.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(stream))
            submission.childNodes[-1].appendChild(x.createElement("name"))
            submission.childNodes[-1].childNodes[-1].appendChild(x.createTextNode(name))
        if s.command:
            submission.appendChild(x.createElement("command"))
            submission.childNodes[-1].appendChild(x.createTextNode(s.command))
        for outfile in s.global_outfiles:
            submission.appendChild(x.createElement("output"))
            submission.childNodes[-1].appendChild(x.createTextNode(outfile))
        submission.appendChild(x.createElement("outds"))
        submission.childNodes[-1].appendChild(x.createTextNode(s.outds))
        for job in s.jobs:
            submission.appendChild(job.to_dom())
        return submission

    def check(s):
        """checks that all output files have unique qualifiers"""
        quals = []
        for j in s.jobs:
            quals += j.outputs_list(True)
        if len(list(set(quals))) != len(quals):
            print("ERROR: found non-unique output file names across the jobs")
            print("(you likely need to review xml options with prepend=true)")
            sys.exit(0)

    def input_datasets(s):
        """returns a list of all used input datasets"""
        DSs = set()
        for j in s.jobs:
            for ds in j.infiles.keys():
                DSs.add(ds)
        return list(DSs)

    def outDS(s):
        """output dataset"""
        return s.outds

    def inDS(s):
        """chooses a dataset we'll call inDS; others will become secondaryDS"""
        # user manually labeled one of datasets as primary, so make it inDS:
        if s.primaryds:
            return s.primaryds
        # OR: choose inDS dataset randomly
        else:
            return s.input_datasets()[0]

    def secondaryDSs(s):
        """returns all secondaryDSs. This excludes inDS, unless inDS is managed by prun"""
        return [d for d in s.input_datasets() if d != s.inDS()]

    def secondaryDSs_config(s, filter=True):
        """returns secondaryDSs string in prun format:
        StreamName:nFilesPerJob:DatasetName[:MatchingPattern[:nSkipFiles]]
        nFilesPerJob is set to zero, so that it is updated later from actual file count.
        MatchingPattern is an OR-separated list of actual file names.
        """
        out = []
        DSs = s.secondaryDSs()
        for i, DS in enumerate(DSs):
            if DS in s.inds:
                stream = s.inds[DS]
            else:
                stream = "IN%d" % (i + 1,)
            if filter:
                out.append(f"{stream}:0:{DS}:{s.files_in_DS(DS, regex=True)}")
            else:
                out.append(f"{stream}:0:{DS}")
        return ",".join(out)

    def writeInputToTxt(s):
        """Prepares prun option --writeInputToTxt
        comma-separated list of STREAM:STREAM.files.dat
        """
        out = []
        DSs = s.secondaryDSs()
        for i, DS in enumerate(DSs):
            if DS in s.inds:
                stream = s.inds[DS]
            else:
                stream = "IN%d" % (i + 1,)
            out.append(f"{stream}:{stream}.files.dat")
        out.append("IN:IN.files.dat")
        return ",".join(out)

    def files_in_DS(s, DS, regex=False):
        """Returns a list of all files from a given dataset
        that will be used in at least one job in this submission
        If regex==True, the list is converted to a regex string
        """
        assert DS in s.input_datasets(), f"ERROR: dataset {DS} was not requested in the xml file"
        files = []
        for j in s.jobs:
            if DS in j.infiles.keys():
                files += j.infiles[DS]
        if regex:
            return "|".join(sorted(list(set(files))))
        else:
            return sorted(list(set(files)))

    def nFiles_in_DS(s, DS):
        return len(s.files_in_DS(DS))

    def nJobs(s):
        return len(s.jobs)

    def dump(s, verbose=True):
        """prints a summary of this submission"""

        def P(key, value=""):
            if value == "":
                print(key)
            else:
                print((key + ":").ljust(14) + " " + value)

        P("XML FILE LOADED", s.fname)
        P("Title", s.title)
        P("Command", s.command)
        P("InDS", s.inDS())
        P("Output DS", s.outds)
        P("njobs", s.nJobs())
        if verbose:
            for i, job in enumerate(s.jobs):
                P("===============> JOB%d" % i)
                P("command", job.exec_string())
                P("outfiles", job.outputs())
                P("INPUTS:")
                j = 0
                for dsname, files in job.infiles.items():
                    P("  Dataset%d" % j, dsname)
                    for k, fname in enumerate(files):
                        P("     File%d" % k, fname)
                    j += 1


if __name__ == "__main__":
    p = dom_parser("./job.xml")
    p.dump()



import uuid
from urllib.parse import quote

from userinterface import Client

taskParamMap = {}

taskParamMap["noInput"] = True
taskParamMap["nEventsPerJob"] = 1
taskParamMap["nEvents"] = 5
taskParamMap["taskName"] = str(uuid.uuid4())
taskParamMap["userName"] = "pandasrv1"
taskParamMap["vo"] = "atlas"
taskParamMap["taskPriority"] = 1000
taskParamMap["reqID"] = 12345
taskParamMap["architecture"] = ""
taskParamMap["hpoWorkflow"] = True
taskParamMap["transUses"] = ""
taskParamMap["transHome"] = ""
taskParamMap["transPath"] = "http://pandaserver.cern.ch:25080/trf/user/runHPO-00-00-01"
taskParamMap["processingType"] = "simul"
taskParamMap["prodSourceLabel"] = "test"
taskParamMap["useLocalIO"] = 1
taskParamMap["taskType"] = "prod"
taskParamMap["workingGroup"] = "AP_HPO"
taskParamMap["coreCount"] = 1
taskParamMap["site"] = "BNL_PROD_UCORE"
taskParamMap["nucleus"] = "CERN-PROD"
taskParamMap["cloud"] = "WORLD"

logDatasetName = f"panda.jeditest.log.{uuid.uuid4()}"
outDatasetName = f"panda.jeditest.HPO.{uuid.uuid4()}"

taskParamMap["log"] = {
    "dataset": logDatasetName,
    "type": "template",
    "param_type": "log",
    "token": "ddd:.*DATADISK",
    "destination": "(type=DATADISK)\(dontkeeplog=True)",
    "offset": 1000,
    "value": f"{logDatasetName}.${{SN}}.log.tgz",
}

taskParamMap["hpoRequestData"] = {
    "sandbox": "gitlab-registry.cern.ch/zhangruihpc/steeringcontainer:latest",
    "executable": "docker",
    "arguments": '/bin/bash -c "hpogrid generate --n_point=%NUM_POINTS ' "--max_point=%MAX_POINTS --infile=$PWD/%IN  --outfile=$PWD/%OUT " '-l=nevergrad"',
    "output_json": "output.json",
    "max_points": 10,
    "num_points_per_generation": 2,
}

taskParamMap["container_name"] = "docker://gitlab-registry.cern.ch/zhangruihpc/evaluationcontainer:mlflow"

taskParamMap["jobParameters"] = [
    {"type": "constant", "value": f"-o out.json -j \"\" -p \"{quote('bash ./exec_in_container.sh')}\""},
    {"type": "constant", "value": "--writeInputToTxt IN_DATA:input_ds.json --inSampleFile input_sample.json"},
    {"type": "constant", "value": "-a aaa.tgz --sourceURL https://aipanda048.cern.ch:25443"},
    {"type": "constant", "value": "--inMap \"{'IN_DATA': ${IN_DATA/T}}\""},
    {
        "type": "template",
        "param_type": "input",
        "value": '-i "${IN_DATA/T}"',
        "dataset": "mc16_13TeV.501103.MGPy8EG_StauStauDirect_220p0_1p0_TFilt.merge.EVNT.e8102_e7400_tid21342682_00",
        "attributes": "nosplit,repeat",
    },
    {
        "type": "template",
        "param_type": "output",
        "token": "ATLASDATADISK",
        "value": "$JEDITASKID.metrics.${SN}.tgz",
        "dataset": outDatasetName,
        "hidden": True,
    },
    {
        "type": "constant",
        "value": "--outMetricsFile=${OUTPUT0}^metrics.tgz",
    },
]

print(Client.insertTaskParams(taskParamMap))



import os
import socket
import sys
import traceback

from pandajedi.jedicore.JediTaskSpec import JediTaskSpec


# watchdog to take actions for jumbo jobs
class JumboWatchDog:
    # constructor
    def __init__(self, taskBufferIF, ddmIF, log, vo, prodSourceLabel):
        self.taskBufferIF = taskBufferIF
        self.ddmIF = ddmIF
        self.pid = f"{socket.getfqdn().split('.')[0]}-{os.getpid()}_{os.getpgrp()}-jumbo"
        self.log = log
        self.vo = vo
        self.prodSourceLabel = prodSourceLabel
        self.component = "JumboWatchDog"
        self.dryRun = True

    # main
    def run(self):
        try:
            # get process lock
            locked = self.taskBufferIF.lockProcess_JEDI(
                vo=self.vo,
                prodSourceLabel=self.prodSourceLabel,
                cloud=None,
                workqueue_id=None,
                resource_name=None,
                component=self.component,
                pid=self.pid,
                timeLimit=10,
            )
            if not locked:
                self.log.debug(f"component={self.component} skipped since locked by another")
                return
            # get parameters for conversion
            self.log.debug(f"component={self.component} start")
            maxTasks = self.taskBufferIF.getConfigValue(self.component, "JUMBO_MAX_TASKS", "jedi", self.vo)
            if maxTasks is None:
                maxTasks = 1
            nEventsToDisable = self.taskBufferIF.getConfigValue(self.component, "JUMBO_MIN_EVENTS_DISABLE", "jedi", self.vo)
            if nEventsToDisable is None:
                nEventsToDisable = 100000
            nEventsToEnable = self.taskBufferIF.getConfigValue(self.component, "JUMBO_MIN_EVENTS_ENABLE", "jedi", self.vo)
            if nEventsToEnable is None:
                nEventsToEnable = nEventsToDisable * 10
            maxEvents = self.taskBufferIF.getConfigValue(self.component, "JUMBO_MAX_EVENTS", "jedi", self.vo)
            if maxEvents is None:
                maxEvents = maxTasks * nEventsToEnable // 2
            nJumboPerTask = self.taskBufferIF.getConfigValue(self.component, "JUMBO_PER_TASK", "jedi", self.vo)
            if nJumboPerTask is None:
                nJumboPerTask = 1
            nJumboPerSite = self.taskBufferIF.getConfigValue(self.component, "JUMBO_PER_SITE", "jedi", self.vo)
            if nJumboPerSite is None:
                nJumboPerSite = 1
            maxPrio = self.taskBufferIF.getConfigValue(self.component, "JUMBO_MAX_CURR_PRIO", "jedi", self.vo)
            if maxPrio is None:
                maxPrio = 500
            progressToBoost = self.taskBufferIF.getConfigValue(self.component, "JUMBO_PROG_TO_BOOST", "jedi", self.vo)
            if progressToBoost is None:
                progressToBoost = 95
            maxFilesToBoost = self.taskBufferIF.getConfigValue(self.component, "JUMBO_MAX_FILES_TO_BOOST", "jedi", self.vo)
            if maxFilesToBoost is None:
                maxFilesToBoost = 500
            prioToBoost = 900
            prioWhenDisabled = self.taskBufferIF.getConfigValue(self.component, "JUMBO_PRIO_DISABLED", "jedi", self.vo)
            if prioWhenDisabled is None:
                prioWhenDisabled = 500
            # get current info
            tasksWithJumbo = self.taskBufferIF.getTaskWithJumbo_JEDI(self.vo, self.prodSourceLabel)
            totEvents = 0
            doneEvents = 0
            nTasks = 0
            for jediTaskID, taskData in tasksWithJumbo.items():
                # disable jumbo
                if taskData["useJumbo"] != JediTaskSpec.enum_useJumbo["disabled"] and taskData["site"] is None:
                    if taskData["nEvents"] - taskData["nEventsDone"] < nEventsToDisable:
                        # disable
                        self.log.info(
                            "component={0} disable jumbo in jediTaskID={1} due to n_events_to_process={2} < {3}".format(
                                self.component, jediTaskID, taskData["nEvents"] - taskData["nEventsDone"], nEventsToDisable
                            )
                        )
                        self.taskBufferIF.enableJumboJobs(jediTaskID, 0, 0)
                    else:
                        # wait
                        nTasks += 1
                        totEvents += taskData["nEvents"]
                        doneEvents += taskData["nEventsDone"]
                        self.log.info(
                            "component={0} keep jumbo in jediTaskID={1} due to n_events_to_process={2} > {3}".format(
                                self.component, jediTaskID, taskData["nEvents"] - taskData["nEventsDone"], nEventsToDisable
                            )
                        )
                # increase priority for jumbo disabled
                if taskData["useJumbo"] == JediTaskSpec.enum_useJumbo["disabled"] and taskData["currentPriority"] < prioWhenDisabled:
                    self.taskBufferIF.changeTaskPriorityPanda(jediTaskID, prioWhenDisabled)
                    self.log.info(f"component={self.component} priority boost to {prioWhenDisabled} after disabing jumbo in in jediTaskID={jediTaskID}")
                # increase priority when close to completion
                if (
                    taskData["nEvents"] > 0
                    and (taskData["nEvents"] - taskData["nEventsDone"]) * 100 // taskData["nEvents"] < progressToBoost
                    and taskData["currentPriority"] < prioToBoost
                    and (taskData["nFiles"] - taskData["nFilesDone"]) < maxFilesToBoost
                ):
                    # boost
                    tmpStr = "component={0} priority boost to {5} for jediTaskID={1} due to n_events_done={2} > {3}*{4}% ".format(
                        self.component, jediTaskID, taskData["nEventsDone"], taskData["nEvents"], progressToBoost, prioToBoost
                    )
                    tmpStr += f"n_files_remaining={taskData['nFiles'] - taskData['nFilesDone']} < {maxFilesToBoost}"
                    self.log.info(tmpStr)
                    self.taskBufferIF.changeTaskPriorityPanda(jediTaskID, prioToBoost)
                # kick pending
                if taskData["taskStatus"] in ["pending", "running"] and taskData["useJumbo"] in [
                    JediTaskSpec.enum_useJumbo["pending"],
                    JediTaskSpec.enum_useJumbo["running"],
                ]:
                    nActiveJumbo = 0
                    for computingSite, jobStatusMap in taskData["jumboJobs"].items():
                        for jobStatus, nJobs in jobStatusMap.items():
                            if jobStatus in ["defined", "assigned", "activated", "sent", "starting", "running", "transferring", "holding"]:
                                nActiveJumbo += nJobs
                    if nActiveJumbo == 0:
                        self.log.info(f"component={self.component} kick jumbo in {taskData['taskStatus']} jediTaskID={jediTaskID}")
                        self.taskBufferIF.kickPendingTasksWithJumbo_JEDI(jediTaskID)
                # reset input to re-generate co-jumbo
                if taskData["currentPriority"] >= prioToBoost:
                    nReset = self.taskBufferIF.resetInputToReGenCoJumbo_JEDI(jediTaskID)
                    if nReset is not None and nReset > 0:
                        self.log.info(f"component={self.component} reset {nReset} inputs to regenerate co-jumbo for jediTaskID={jediTaskID}")
                    else:
                        self.log.debug(f"component={self.component} tried to reset inputs to regenerate co-jumbo with {nReset} for jediTaskID={jediTaskID}")
            self.log.info(
                f"component={self.component} total_events={totEvents} n_events_to_process={totEvents - doneEvents} n_tasks={nTasks} available for jumbo"
            )

            self.log.debug(f"component={self.component} done")
        except Exception:
            # error
            errtype, errvalue = sys.exc_info()[:2]
            errStr = f": {errtype.__name__} {errvalue}"
            errStr.strip()
            errStr += traceback.format_exc()
            self.log.error(errStr)



jedi_task_plain = """
Subject: JEDI notification for TaskID:{jedi_task_id} ({n_succeeded}/{n_total} {msg_succeeded})

Summary of TaskID:{jedi_task_id}

Created : {creation_time} (UTC)
Ended   : {end_time} (UTC)

Final Status : {task_status}

Total Number of {input_str}   : {n_total}
             Succeeded   : {n_succeeded}
             Failed      : {n_failed}
             {cancelled_str} : {n_cancelled}


Error Dialog : {error_dialog}

{dataset_summary}

Parameters : {command}


PandaMonURL : https://bigpanda.cern.ch/task/{jedi_task_id}/


Estimated carbon footprint for the task
     Succeeded   : {carbon_succeeded}
     Failed      : {carbon_failed}
     Cancelled   : {carbon_cancelled}
     _________________
     Total       : {carbon_total}
(More details on estimation: https://panda-wms.readthedocs.io/en/latest/advanced/carbon_footprint.html)


Report Panda problems of any sort to

  the eGroup for help request
    hn-atlas-dist-analysis-help@cern.ch

  the JIRA portal for software bug
    https://its.cern.ch/jira/browse/ATLASPANDA
"""

html_head = """
<!DOCTYPE html>
<html>
<head>
    <style>
        body {{
            font-family: Arial, sans-serif;
        }}
        table {{
            width: 100%;
            border-collapse: collapse;
            margin-top: 10px;
        }}
        th, td {{
            border: 1px solid #dddddd;
            text-align: left;
            padding: 8px;
        }}
        th {{
            background-color: #f2f2f2;
        }}
        .logo {{
            width: 100px; /* Adjust the size as needed */
        }}
    </style>
    <title> {title} </title>
</head>
"""

jedi_task_html_body = """
<body>
    <img class="logo" src="https://panda-wms.readthedocs.io/en/latest/_static/PanDA-rev-logo-small-200px.jpg" alt="PanDA Logo">
    <h2>Summary of TaskID: <a href="https://bigpanda.cern.ch/task/{jedi_task_id}">{jedi_task_id}</a> </h2>
    <table>
        <tr>
            <th>Detail</th>
            <th>Value</th>
        </tr>
        <tr>
            <td>Created</td>
            <td>{creation_time}</td>
        </tr>
        <tr>
            <td>Ended</td>
            <td>{end_time}</td>
        </tr>
        <tr>
            <td>Final Status</td>
            <td>{task_status}</td>
        </tr>
    </table>

    <h3>Total Number of {input_str}</h3>
    <table>
        <tr>
            <th>Category</th>
            <th>Count</th>
        </tr>
        <tr>
            <td>Succeeded</td>
            <td>{n_succeeded}</td>
        </tr>
        <tr>
            <td>Failed</td>
            <td>{n_failed}</td>
        </tr>
        <tr>
            <td>{cancelled_str}</td>
            <td>{n_cancelled}</td>
        </tr>
        <tr>
            <td>Total</td>
            <td>{n_total}</td>
        </tr>
    </table>

    <p><strong>Error Dialog:</strong> {error_dialog}</p>

    <h3>Datasets</h3>
    <table>
        <tr>
            <td><strong>In</strong></td>
            <td>{datasets_in}</td>
        </tr>
        <tr>
            <td><strong>Out</strong></td>
            <td>{datasets_out}</td>
        </tr>
        <tr>
            <td><strong>Log</strong></td>
            <td>{datasets_log}</td>
        </tr>
    </table>

    <h3>Parameters</h3>
    <table>
        <tr>
            <td><strong>Command</strong></td>
            <td>{command}</td>
        </tr>
    </table>

    <h3>Estimated Carbon Footprint for the Task</h3>
    <table>
        <tr>
            <th>Category</th>
            <th>gCO2</th>
        </tr>
        <tr>
            <td>Succeeded</td>
            <td>{carbon_succeeded}</td>
        </tr>
        <tr>
            <td>Failed</td>
            <td>{carbon_failed}</td>
        </tr>
        <tr>
            <td>Cancelled</td>
            <td>{carbon_cancelled}</td>
        </tr>
        <tr>
            <td><strong>Total</strong></td>
            <td><strong>{carbon_total}</strong></td>
        </tr>
    </table>
    <p>More details on estimation: <a href="https://panda-wms.readthedocs.io/en/latest/advanced/carbon_footprint.html">https://panda-wms.readthedocs.io/en/latest/advanced/carbon_footprint.html</a></p>

    <h3>Panda Problem Reporting</h3>
    <ul>
        <li>The eGroup for help request: <a href="mailto:hn-atlas-dist-analysis-help@cern.ch">hn-atlas-dist-analysis-help@cern.ch</a></li>
        <li>The JIRA portal for software bug: <a href="https://its.cern.ch/jira/browse/ATLASPANDA">https://its.cern.ch/jira/browse/ATLASPANDA</a></li>
    </ul>
</body>
</html>
"""



import datetime
import math
import os
import re
import socket
import sys
import time
import traceback
import uuid

from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.ThreadUtils import ListWithLock, ThreadPool, WorkerThread
from pandajedi.jedirefine import RefinerUtils

from .JediKnight import JediKnight

try:
    import idds.common.constants
    import idds.common.utils
    from idds.client.client import Client as iDDS_Client
except ImportError:
    pass


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# worker class to take care of DatasetContents table
class ContentsFeeder(JediKnight):
    # constructor
    def __init__(self, commuChannel, taskBufferIF, ddmIF, vos, prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        self.pid = f"{socket.getfqdn().split('.')[0]}-{os.getpid()}_{os.getpgrp()}-con"
        JediKnight.__init__(self, commuChannel, taskBufferIF, ddmIF, logger)

    # main
    def start(self):
        # start base class
        JediKnight.start(self)
        # go into main loop
        while True:
            startTime = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
            try:
                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # get the list of datasets to feed contents to DB
                        tmpList = self.taskBufferIF.getDatasetsToFeedContents_JEDI(vo, prodSourceLabel)
                        if tmpList is None:
                            # failed
                            logger.error("failed to get the list of datasets to feed contents")
                        else:
                            logger.debug(f"got {len(tmpList)} datasets")
                            # put to a locked list
                            dsList = ListWithLock(tmpList)
                            # make thread pool
                            threadPool = ThreadPool()
                            # make workers
                            nWorker = jedi_config.confeeder.nWorkers
                            for iWorker in range(nWorker):
                                thr = ContentsFeederThread(dsList, threadPool, self.taskBufferIF, self.ddmIF, self.pid)
                                thr.start()
                            # join
                            threadPool.join()
            except Exception:
                errtype, errvalue = sys.exc_info()[:2]
                logger.error(f"failed in {self.__class__.__name__}.start() with {errtype.__name__} {errvalue}")
            # sleep if needed
            loopCycle = jedi_config.confeeder.loopCycle
            timeDelta = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None) - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)
            # randomize cycle
            self.randomSleep(max_val=loopCycle)


# thread for real worker
class ContentsFeederThread(WorkerThread):
    # constructor
    def __init__(self, taskDsList, threadPool, taskbufferIF, ddmIF, pid):
        # initialize woker with no semaphore
        WorkerThread.__init__(self, None, threadPool, logger)
        # attributres
        self.taskDsList = taskDsList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF
        self.msgType = "contentsfeeder"
        self.pid = pid

    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 10
                taskDsList = self.taskDsList.get(nTasks)
                # no more datasets
                if len(taskDsList) == 0:
                    self.logger.debug(f"{self.__class__.__name__} terminating since no more items")
                    return
                # feed to tasks
                self.feed_contents_to_tasks(taskDsList)
            except Exception as e:
                logger.error(f"{self.__class__.__name__} failed in runImpl() with {str(e)}: {traceback.format_exc()}")

    # feed contents to tasks
    def feed_contents_to_tasks(self, task_ds_list):
        # max number of file records per dataset
        maxFileRecords = 200000
        # loop over all tasks
        for jediTaskID, dsList in task_ds_list:
            allUpdated = True
            taskBroken = False
            taskOnHold = False
            runningTask = False
            taskToFinish = False
            missingMap = {}
            datasetsIdxConsistency = []

            # get task
            tmpStat, taskSpec = self.taskBufferIF.getTaskWithID_JEDI(jediTaskID, False, True, self.pid, 10, clearError=True)
            if not tmpStat or taskSpec is None:
                self.logger.debug(f"failed to get taskSpec for jediTaskID={jediTaskID}")
                continue

            # make logger
            try:
                gshare = "_".join(taskSpec.gshare.split(" "))
            except Exception:
                gshare = "Undefined"
            tmpLog = MsgWrapper(self.logger, f"<jediTaskID={jediTaskID} gshare={gshare}>")

            try:
                # get task parameters
                taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(jediTaskID)
                taskParamMap = RefinerUtils.decodeJSON(taskParam)
            except Exception as e:
                tmpLog.error(f"task param conversion from json failed with {str(e)}")
                # unlock
                tmpStat = self.taskBufferIF.unlockSingleTask_JEDI(jediTaskID, self.pid)
                tmpLog.debug(f"unlocked with {tmpStat}")
                continue
            # renaming of parameters
            if "nEventsPerInputFile" in taskParamMap:
                taskParamMap["nEventsPerFile"] = taskParamMap["nEventsPerInputFile"]
            # the number of files per job
            nFilesPerJob = taskSpec.getNumFilesPerJob()
            # the number of chunks used by scout
            nChunksForScout = 10
            # load XML
            if taskSpec.useLoadXML():
                xmlConfig = taskParamMap["loadXML"]
            else:
                xmlConfig = None
            # skip files used by another task
            if "skipFilesUsedBy" in taskParamMap:
                skipFilesUsedBy = taskParamMap["skipFilesUsedBy"]
            else:
                skipFilesUsedBy = None
            # check no wait
            noWaitParent = False
            parentOutDatasets = set()
            if taskSpec.noWaitParent() and taskSpec.parent_tid not in [None, taskSpec.jediTaskID]:
                tmpStat = self.taskBufferIF.checkParentTask_JEDI(taskSpec.parent_tid)
                if tmpStat is None or tmpStat == "running":
                    noWaitParent = True
                    # get output datasets from parent task
                    tmpParentStat, tmpParentOutDatasets = self.taskBufferIF.getDatasetsWithJediTaskID_JEDI(taskSpec.parent_tid, ["output", "log"])
                    # collect dataset names
                    for tmpParentOutDataset in tmpParentOutDatasets:
                        parentOutDatasets.add(tmpParentOutDataset.datasetName)
                        if tmpParentOutDataset.containerName:
                            if tmpParentOutDataset.containerName.endswith("/"):
                                parentOutDatasets.add(tmpParentOutDataset.containerName)
                                parentOutDatasets.add(tmpParentOutDataset.containerName[:-1])
                            else:
                                parentOutDatasets.add(tmpParentOutDataset.containerName)
                                parentOutDatasets.add(tmpParentOutDataset.containerName + "/")
            # loop over all datasets
            nFilesMaster = 0
            checkedMaster = False
            setFrozenTime = True
            if not taskBroken:
                ddmIF = self.ddmIF.getInterface(taskSpec.vo, taskSpec.cloud)
                origNumFiles = None
                if "nFiles" in taskParamMap:
                    origNumFiles = taskParamMap["nFiles"]
                id_to_container = {}
                [id_to_container.update({datasetSpec.datasetID: datasetSpec.containerName}) for datasetSpec in dsList]
                for datasetSpec in dsList:
                    tmpLog.debug(f"start loop for {datasetSpec.datasetName}(id={datasetSpec.datasetID})")
                    # index consistency
                    if datasetSpec.indexConsistent():
                        datasetsIdxConsistency.append(datasetSpec.datasetID)
                    # prestaging
                    if taskSpec.inputPreStaging() and (datasetSpec.isMaster() or datasetSpec.isSeqNumber()):
                        nStaging = self.taskBufferIF.getNumStagingFiles_JEDI(taskSpec.jediTaskID)
                        if nStaging is not None and nStaging == 0 and datasetSpec.nFiles > 0:
                            inputPreStaging = False
                        else:
                            inputPreStaging = True
                    else:
                        inputPreStaging = False
                    # get dataset metadata
                    tmpLog.debug("get metadata")
                    gotMetadata = False
                    stateUpdateTime = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
                    try:
                        if not datasetSpec.isPseudo():
                            tmpMetadata = ddmIF.getDatasetMetaData(datasetSpec.datasetName, ignore_missing=True)
                        else:
                            # dummy metadata for pseudo dataset
                            tmpMetadata = {"state": "closed"}
                        # set mutable when the dataset is open and parent is running or task is configured to run until the dataset is closed
                        if (noWaitParent or taskSpec.runUntilClosed() or inputPreStaging) and (
                            tmpMetadata["state"] == "open"
                            or datasetSpec.datasetName in parentOutDatasets
                            or datasetSpec.datasetName.split(":")[-1] in parentOutDatasets
                            or inputPreStaging
                        ):
                            # dummy metadata when parent is running
                            tmpMetadata = {"state": "mutable"}
                        gotMetadata = True
                    except Exception:
                        errtype, errvalue = sys.exc_info()[:2]
                        tmpLog.error(f"{self.__class__.__name__} failed to get metadata to {errtype.__name__}:{errvalue}")
                        if errtype == Interaction.JEDIFatalError:
                            # fatal error
                            datasetStatus = "broken"
                            taskBroken = True
                            # update dataset status
                            self.updateDatasetStatus(datasetSpec, datasetStatus, tmpLog)
                        else:
                            if not taskSpec.ignoreMissingInDS():
                                # temporary error
                                taskOnHold = True
                            else:
                                # ignore missing
                                datasetStatus = "failed"
                                # update dataset status
                                self.updateDatasetStatus(datasetSpec, datasetStatus, tmpLog)
                        taskSpec.setErrDiag(f"failed to get metadata for {datasetSpec.datasetName}")
                        if not taskSpec.ignoreMissingInDS():
                            allUpdated = False
                    else:
                        # to skip missing dataset
                        if tmpMetadata["state"] == "missing":
                            # ignore missing
                            datasetStatus = "finished"
                            # update dataset status
                            self.updateDatasetStatus(datasetSpec, datasetStatus, tmpLog, "closed")
                            tmpLog.debug(f"disabled missing {datasetSpec.datasetName}")
                            continue
                        # get file list specified in task parameters
                        if taskSpec.is_work_segmented() and not datasetSpec.isPseudo() and not datasetSpec.isMaster():
                            fileList = []
                            includePatt = []
                            excludePatt = []
                            try:
                                segment_id = int(id_to_container[datasetSpec.masterID].split("/")[-1])
                                for item in taskParamMap["segmentSpecs"]:
                                    if item["id"] == segment_id:
                                        if "files" in item:
                                            fileList = item["files"]
                                        elif "datasets" in item:
                                            for tmpDatasetName in item["datasets"]:
                                                tmpRet = ddmIF.getFilesInDataset(tmpDatasetName)
                                                fileList += [tmpAttr["lfn"] for tmpAttr in tmpRet.values()]
                            except Exception:
                                pass
                        else:
                            fileList, includePatt, excludePatt = RefinerUtils.extractFileList(taskParamMap, datasetSpec.datasetName)
                        # get the number of events in metadata
                        if "getNumEventsInMetadata" in taskParamMap:
                            getNumEvents = True
                        else:
                            getNumEvents = False
                        # get file list from DDM
                        tmpLog.debug("get files")
                        try:
                            useInFilesWithNewAttemptNr = False
                            skipDuplicate = not datasetSpec.useDuplicatedFiles()
                            if not datasetSpec.isPseudo():
                                if fileList != [] and "useInFilesInContainer" in taskParamMap and datasetSpec.containerName not in ["", None]:
                                    # read files from container if file list is specified in task parameters
                                    tmpDatasetName = datasetSpec.containerName
                                else:
                                    tmpDatasetName = datasetSpec.datasetName
                                # use long format for LB
                                longFormat = False
                                if taskSpec.respectLumiblock() or taskSpec.orderByLB():
                                    longFormat = True
                                tmpRet = ddmIF.getFilesInDataset(tmpDatasetName, getNumEvents=getNumEvents, skipDuplicate=skipDuplicate, longFormat=longFormat)
                                tmpLog.debug(f"got {len(tmpRet)} files in {tmpDatasetName}")
                                # remove lost files
                                """
                                tmpLostFiles = ddmIF.findLostFiles(tmpDatasetName,tmpRet)
                                if tmpLostFiles != {}:
                                    tmpLog.debug('found {0} lost files in {1}'.format(len(tmpLostFiles),tmpDatasetName))
                                    for tmpListGUID,tmpLostLFN in tmpLostFiles.items():
                                        tmpLog.debug('removed {0}'.format(tmpLostLFN))
                                        del tmpRet[tmpListGUID]
                                """
                            else:
                                if datasetSpec.isSeqNumber():
                                    # make dummy files for seq_number
                                    if datasetSpec.getNumRecords() is not None:
                                        nPFN = datasetSpec.getNumRecords()
                                    elif origNumFiles is not None:
                                        nPFN = origNumFiles
                                        if "nEventsPerFile" in taskParamMap and taskSpec.get_min_granularity():
                                            nPFN = nPFN * taskParamMap["nEventsPerFile"] // taskSpec.get_min_granularity()
                                        elif (
                                            "nEventsPerJob" in taskParamMap
                                            and "nEventsPerFile" in taskParamMap
                                            and taskParamMap["nEventsPerFile"] > taskParamMap["nEventsPerJob"]
                                        ):
                                            nPFN = nPFN * math.ceil(taskParamMap["nEventsPerFile"] / taskParamMap["nEventsPerJob"])
                                        elif "nEventsPerJob" in taskParamMap and "nEventsPerFile" not in taskParamMap:
                                            max_events_in_file = self.taskBufferIF.get_max_events_in_dataset(jediTaskID, datasetSpec.masterID)
                                            if max_events_in_file and max_events_in_file > taskParamMap["nEventsPerJob"]:
                                                nPFN = nPFN * math.ceil(max_events_in_file / taskParamMap["nEventsPerJob"])
                                    elif "nEvents" in taskParamMap and "nEventsPerJob" in taskParamMap:
                                        nPFN = math.ceil(taskParamMap["nEvents"] / taskParamMap["nEventsPerJob"])
                                    elif "nEvents" in taskParamMap and "nEventsPerFile" in taskParamMap and taskSpec.getNumFilesPerJob() is not None:
                                        nPFN = math.ceil(taskParamMap["nEvents"] / taskParamMap["nEventsPerFile"] / taskSpec.getNumFilesPerJob())
                                    else:
                                        # the default number of records for seq_number
                                        seqDefNumRecords = 10000
                                        # get nFiles of the master
                                        tmpMasterAtt = self.taskBufferIF.getDatasetAttributes_JEDI(datasetSpec.jediTaskID, datasetSpec.masterID, ["nFiles"])
                                        # use nFiles of the master as the number of records if it is larger than the default
                                        if "nFiles" in tmpMasterAtt and tmpMasterAtt["nFiles"] > seqDefNumRecords:
                                            nPFN = tmpMasterAtt["nFiles"]
                                        else:
                                            nPFN = seqDefNumRecords
                                        # check usedBy
                                        if skipFilesUsedBy is not None:
                                            for tmpJediTaskID in str(skipFilesUsedBy).split(","):
                                                tmpParentAtt = self.taskBufferIF.getDatasetAttributesWithMap_JEDI(
                                                    tmpJediTaskID, {"datasetName": datasetSpec.datasetName}, ["nFiles"]
                                                )
                                                if "nFiles" in tmpParentAtt and tmpParentAtt["nFiles"]:
                                                    nPFN += tmpParentAtt["nFiles"]
                                    if nPFN > maxFileRecords:
                                        raise Interaction.JEDIFatalError(f"Too many file records for seq_number >{maxFileRecords}")
                                    tmpRet = {}
                                    # get offset
                                    tmpOffset = datasetSpec.getOffset()
                                    tmpOffset += 1
                                    for iPFN in range(nPFN):
                                        tmpRet[str(uuid.uuid4())] = {
                                            "lfn": iPFN + tmpOffset,
                                            "scope": None,
                                            "filesize": 0,
                                            "checksum": None,
                                        }
                                elif not taskSpec.useListPFN():
                                    # dummy file list for pseudo dataset
                                    tmpRet = {
                                        str(uuid.uuid4()): {
                                            "lfn": "pseudo_lfn",
                                            "scope": None,
                                            "filesize": 0,
                                            "checksum": None,
                                        }
                                    }
                                else:
                                    # make dummy file list for PFN list
                                    if "nFiles" in taskParamMap:
                                        nPFN = taskParamMap["nFiles"]
                                    else:
                                        nPFN = 1
                                    tmpRet = {}
                                    for iPFN in range(nPFN):
                                        base_name = taskParamMap["pfnList"][iPFN].split("/")[-1]
                                        n_events = None
                                        if "^" in base_name:
                                            base_name, n_events = base_name.split("^")

                                        tmpRet[str(uuid.uuid4())] = {
                                            "lfn": f"{iPFN:06d}:{base_name}",
                                            "scope": None,
                                            "filesize": 0,
                                            "checksum": None,
                                            "events": n_events,
                                        }
                        except Exception:
                            errtype, errvalue = sys.exc_info()[:2]
                            reason = str(errvalue)
                            tmpLog.error(f"failed to get files in {self.__class__.__name__}:{errtype.__name__} due to {reason}")
                            if errtype == Interaction.JEDIFatalError:
                                # fatal error
                                datasetStatus = "broken"
                                taskBroken = True
                                # update dataset status
                                self.updateDatasetStatus(datasetSpec, datasetStatus, tmpLog)
                            else:
                                # temporary error
                                taskOnHold = True
                            taskSpec.setErrDiag(f"failed to get files for {datasetSpec.datasetName} due to {reason}")
                            allUpdated = False
                        else:
                            # parameters for master input
                            respectLB = False
                            useRealNumEvents = False
                            if datasetSpec.isMaster():
                                # respect LB boundaries
                                respectLB = taskSpec.respectLumiblock()
                                # use real number of events
                                useRealNumEvents = taskSpec.useRealNumEvents()
                            # the number of events per file
                            nEventsPerFile = None
                            nEventsPerJob = None
                            nEventsPerRange = None
                            tgtNumEventsPerJob = None
                            if (datasetSpec.isMaster() and ("nEventsPerFile" in taskParamMap or useRealNumEvents)) or (
                                datasetSpec.isPseudo() and "nEvents" in taskParamMap and not datasetSpec.isSeqNumber()
                            ):
                                if "nEventsPerFile" in taskParamMap:
                                    nEventsPerFile = taskParamMap["nEventsPerFile"]
                                elif datasetSpec.isMaster() and datasetSpec.isPseudo() and "nEvents" in taskParamMap:
                                    # use nEvents as nEventsPerFile for pseudo input
                                    nEventsPerFile = taskParamMap["nEvents"]
                                if taskSpec.get_min_granularity():
                                    nEventsPerRange = taskSpec.get_min_granularity()
                                elif "nEventsPerJob" in taskParamMap:
                                    nEventsPerJob = taskParamMap["nEventsPerJob"]
                                if "tgtNumEventsPerJob" in taskParamMap:
                                    tgtNumEventsPerJob = taskParamMap["tgtNumEventsPerJob"]
                                    # reset nEventsPerJob
                                    nEventsPerJob = None
                            # max attempts
                            maxAttempt = None
                            maxFailure = None
                            if datasetSpec.isMaster() or datasetSpec.toKeepTrack():
                                # max attempts
                                if taskSpec.disableAutoRetry():
                                    # disable auto retry
                                    maxAttempt = 1
                                elif "maxAttempt" in taskParamMap:
                                    maxAttempt = taskParamMap["maxAttempt"]
                                else:
                                    # use default value
                                    maxAttempt = 3
                                # max failure
                                if "maxFailure" in taskParamMap:
                                    maxFailure = taskParamMap["maxFailure"]
                            # first event number
                            firstEventNumber = None
                            if datasetSpec.isMaster():
                                # first event number
                                firstEventNumber = 1 + taskSpec.getFirstEventOffset()
                            # nMaxEvents
                            nMaxEvents = None
                            if datasetSpec.isMaster() and "nEvents" in taskParamMap:
                                nMaxEvents = taskParamMap["nEvents"]
                            # nMaxFiles
                            nMaxFiles = None
                            if "nFiles" in taskParamMap:
                                if datasetSpec.isMaster():
                                    nMaxFiles = taskParamMap["nFiles"]
                                else:
                                    # calculate for secondary
                                    if not datasetSpec.isPseudo():
                                        # check nFilesPerJob
                                        nFilesPerJobSec = datasetSpec.getNumFilesPerJob()
                                        if nFilesPerJobSec is not None:
                                            nMaxFiles = origNumFiles * nFilesPerJobSec
                                    # check ratio
                                    if nMaxFiles is None:
                                        nMaxFiles = datasetSpec.getNumMultByRatio(origNumFiles)
                                    # multiplied by the number of jobs per file for event-level splitting
                                    if nMaxFiles is not None:
                                        if "nEventsPerFile" in taskParamMap:
                                            if taskSpec.get_min_granularity():
                                                if taskParamMap["nEventsPerFile"] > taskSpec.get_min_granularity():
                                                    nMaxFiles *= float(taskParamMap["nEventsPerFile"]) / float(taskSpec.get_min_granularity())
                                                    nMaxFiles = int(math.ceil(nMaxFiles))
                                            elif "nEventsPerJob" in taskParamMap:
                                                if taskParamMap["nEventsPerFile"] > taskParamMap["nEventsPerJob"]:
                                                    nMaxFiles *= float(taskParamMap["nEventsPerFile"]) / float(taskParamMap["nEventsPerJob"])
                                                    nMaxFiles = int(math.ceil(nMaxFiles))
                                        elif "useRealNumEvents" in taskParamMap:
                                            # reset nMaxFiles since it is unknown
                                            nMaxFiles = None
                            # use scout
                            useScout = False
                            if datasetSpec.isMaster() and taskSpec.useScout() and (datasetSpec.status != "toupdate" or not taskSpec.isPostScout()):
                                useScout = True
                            # use files with new attempt numbers
                            useFilesWithNewAttemptNr = False
                            if not datasetSpec.isPseudo() and fileList != [] and "useInFilesWithNewAttemptNr" in taskParamMap:
                                useFilesWithNewAttemptNr = True
                            # ramCount
                            ramCount = 0
                            # skip short input
                            if (
                                datasetSpec.isMaster()
                                and not datasetSpec.isPseudo()
                                and nEventsPerFile is not None
                                and nEventsPerJob is not None
                                and nEventsPerFile >= nEventsPerJob
                                and "skipShortInput" in taskParamMap
                                and taskParamMap["skipShortInput"] is True
                            ):
                                skipShortInput = True
                            else:
                                skipShortInput = False
                            # skip short output
                            if datasetSpec.isMaster() and not datasetSpec.isPseudo() and taskParamMap.get("skipShortOutput"):
                                skip_short_output = True
                            else:
                                skip_short_output = False
                            # order by
                            if taskSpec.order_input_by() and datasetSpec.isMaster() and not datasetSpec.isPseudo():
                                orderBy = taskSpec.order_input_by()
                            else:
                                orderBy = None
                            # feed files to the contents table
                            tmpLog.debug("update contents")
                            retDB, missingFileList, nFilesUnique, diagMap = self.taskBufferIF.insertFilesForDataset_JEDI(
                                datasetSpec,
                                tmpRet,
                                tmpMetadata["state"],
                                stateUpdateTime,
                                nEventsPerFile,
                                nEventsPerJob,
                                maxAttempt,
                                firstEventNumber,
                                nMaxFiles,
                                nMaxEvents,
                                useScout,
                                fileList,
                                useFilesWithNewAttemptNr,
                                nFilesPerJob,
                                nEventsPerRange,
                                nChunksForScout,
                                includePatt,
                                excludePatt,
                                xmlConfig,
                                noWaitParent,
                                taskSpec.parent_tid,
                                self.pid,
                                maxFailure,
                                useRealNumEvents,
                                respectLB,
                                tgtNumEventsPerJob,
                                skipFilesUsedBy,
                                ramCount,
                                taskSpec,
                                skipShortInput,
                                inputPreStaging,
                                orderBy,
                                maxFileRecords,
                                skip_short_output,
                            )
                            if retDB is False:
                                taskSpec.setErrDiag(f"failed to insert files for {datasetSpec.datasetName}. {diagMap['errMsg']}")
                                allUpdated = False
                                taskBroken = True
                                break
                            elif retDB is None:
                                # the dataset is locked by another or status is not applicable
                                allUpdated = False
                                tmpLog.debug("escape since task or dataset is locked")
                                break
                            elif missingFileList != []:
                                # files are missing
                                tmpErrStr = f"{len(missingFileList)} files missing in {datasetSpec.datasetName}"
                                tmpLog.debug(tmpErrStr)
                                taskSpec.setErrDiag(tmpErrStr)
                                allUpdated = False
                                taskOnHold = True
                                missingMap[datasetSpec.datasetName] = {"datasetSpec": datasetSpec, "missingFiles": missingFileList}
                            else:
                                # reduce the number of files to be read
                                if "nFiles" in taskParamMap:
                                    if datasetSpec.isMaster():
                                        taskParamMap["nFiles"] -= nFilesUnique
                                # reduce the number of files for scout
                                if useScout:
                                    nChunksForScout = diagMap["nChunksForScout"]
                                # number of master input files
                                if datasetSpec.isMaster():
                                    checkedMaster = True
                                    nFilesMaster += nFilesUnique
                            # running task
                            if diagMap["isRunningTask"]:
                                runningTask = True
                            # no activated pending input for noWait
                            if (
                                (noWaitParent or inputPreStaging)
                                and diagMap["nActivatedPending"] == 0
                                and not (useScout and nChunksForScout <= 0)
                                and tmpMetadata["state"] != "closed"
                                and datasetSpec.isMaster()
                            ):
                                tmpErrStr = "insufficient inputs are ready. "
                                tmpErrStr += diagMap["errMsg"]
                                tmpLog.debug(tmpErrStr)
                                taskSpec.setErrDiag(tmpErrStr)
                                taskOnHold = True
                                setFrozenTime = False
                                break
                    tmpLog.debug("end loop")
            # no master input
            if not taskOnHold and not taskBroken and allUpdated and nFilesMaster == 0 and checkedMaster:
                tmpErrStr = "no master input files. input dataset is empty"
                tmpLog.error(tmpErrStr)
                taskSpec.setErrDiag(tmpErrStr, None)
                if noWaitParent:
                    taskOnHold = True
                else:
                    if not taskSpec.allowEmptyInput():
                        taskBroken = True
                    else:
                        taskToFinish = True
            # index consistency
            if not taskOnHold and not taskBroken and len(datasetsIdxConsistency) > 0:
                self.taskBufferIF.removeFilesIndexInconsistent_JEDI(jediTaskID, datasetsIdxConsistency)
            # update task status
            if taskBroken or taskToFinish:
                if taskBroken:
                    # task is broken
                    taskSpec.status = "tobroken"
                else:
                    taskSpec.status = "finishing"
                tmpMsg = f"set task_status={taskSpec.status}"
                tmpLog.info(tmpMsg)
                tmpLog.sendMsg(tmpMsg, self.msgType)
                allRet = self.taskBufferIF.updateTaskStatusByContFeeder_JEDI(jediTaskID, taskSpec, pid=self.pid)
            # change task status unless the task is running
            if not runningTask:
                # send prestaging request
                if taskSpec.inputPreStaging() and taskSpec.is_first_contents_feed():
                    tmpStat, tmpErrStr = self.send_prestaging_request(taskSpec, taskParamMap, dsList, tmpLog)
                    if tmpStat:
                        taskSpec.set_first_contents_feed(False)
                    else:
                        tmpLog.debug(tmpErrStr)
                        taskSpec.setErrDiag(tmpErrStr)
                        taskOnHold = True
                if taskOnHold:
                    # go to pending state
                    if taskSpec.status not in ["broken", "tobroken", "finishing"]:
                        taskSpec.setOnHold()
                    tmpMsg = f"set task_status={taskSpec.status}"
                    tmpLog.info(tmpMsg)
                    tmpLog.sendMsg(tmpMsg, self.msgType)
                    allRet = self.taskBufferIF.updateTaskStatusByContFeeder_JEDI(jediTaskID, taskSpec, pid=self.pid, setFrozenTime=setFrozenTime)
                elif allUpdated:
                    # all OK
                    allRet, newTaskStatus = self.taskBufferIF.updateTaskStatusByContFeeder_JEDI(
                        jediTaskID, getTaskStatus=True, pid=self.pid, useWorldCloud=taskSpec.useWorldCloud()
                    )
                    tmpMsg = f"set task_status={newTaskStatus}"
                    tmpLog.info(tmpMsg)
                    tmpLog.sendMsg(tmpMsg, self.msgType)
                # just unlock
                retUnlock = self.taskBufferIF.unlockSingleTask_JEDI(jediTaskID, self.pid)
                tmpLog.debug(f"unlock not-running task with {retUnlock}")
            else:
                # just unlock
                retUnlock = self.taskBufferIF.unlockSingleTask_JEDI(jediTaskID, self.pid)
                tmpLog.debug(f"unlock task with {retUnlock}")
            # send message to job generator if new inputs are ready
            if not taskOnHold and not taskBroken and allUpdated and taskSpec.is_msg_driven():
                push_ret = self.taskBufferIF.push_task_trigger_message("jedi_job_generator", jediTaskID, task_spec=taskSpec)
                if push_ret:
                    tmpLog.debug("pushed trigger message to jedi_job_generator")
                else:
                    tmpLog.warning("failed to push trigger message to jedi_job_generator")
            tmpLog.debug("done")

    # update dataset
    def updateDatasetStatus(self, datasetSpec, datasetStatus, tmpLog, datasetState=None):
        # update dataset status
        datasetSpec.status = datasetStatus
        datasetSpec.lockedBy = None
        if datasetState:
            datasetSpec.state = datasetState
        tmpLog.info(f"update dataset status to {datasetSpec.status} state to {datasetSpec.state}")
        self.taskBufferIF.updateDataset_JEDI(datasetSpec, {"datasetID": datasetSpec.datasetID, "jediTaskID": datasetSpec.jediTaskID}, lockTask=True)

    # send prestaging request
    def send_prestaging_request(self, task_spec, task_params_map, ds_list, tmp_log):
        try:
            c = iDDS_Client(idds.common.utils.get_rest_host())
            for datasetSpec in ds_list:
                # get rule
                try:
                    tmp_scope, tmp_name = datasetSpec.datasetName.split(":")
                    tmp_name = re.sub("/$", "", tmp_name)
                except Exception:
                    continue
                try:
                    if "prestagingRuleID" in task_params_map:
                        if tmp_name not in task_params_map["prestagingRuleID"]:
                            continue
                        rule_id = task_params_map["prestagingRuleID"][tmp_name]
                    elif "selfPrestagingRule" in task_params_map:
                        if not datasetSpec.isMaster() or datasetSpec.isPseudo():
                            continue
                        rule_id = self.ddmIF.getInterface(task_spec.vo, task_spec.cloud).make_staging_rule(
                            tmp_scope + ":" + tmp_name, task_params_map["selfPrestagingRule"]
                        )
                        if not rule_id:
                            continue
                    else:
                        rule_id = self.ddmIF.getInterface(task_spec.vo, task_spec.cloud).getActiveStagingRule(tmp_scope + ":" + tmp_name)
                        if rule_id is None:
                            continue
                except Exception as e:
                    return False, f"DDM error : {str(e)}"
                # request
                tmp_log.debug(f"sending request to iDDS for {datasetSpec.datasetName}")
                req = {
                    "scope": tmp_scope,
                    "name": tmp_name,
                    "requester": "panda",
                    "request_type": idds.common.constants.RequestType.StageIn,
                    "transform_tag": idds.common.constants.RequestType.StageIn.value,
                    "status": idds.common.constants.RequestStatus.New,
                    "priority": 0,
                    "lifetime": 30,
                    "request_metadata": {
                        "workload_id": task_spec.jediTaskID,
                        "rule_id": rule_id,
                    },
                }
                tmp_log.debug(f"req {str(req)}")
                ret = c.add_request(**req)
                tmp_log.debug(f"got requestID={str(ret)}")
        except Exception as e:
            return False, f"iDDS error : {str(e)}"
        return True, None


# lauch


def launcher(commuChannel, taskBufferIF, ddmIF, vos=None, prodSourceLabels=None):
    p = ContentsFeeder(commuChannel, taskBufferIF, ddmIF, vos, prodSourceLabels)
    p.start()
